
#include "eyd_bitgroup.hh"

namespace EydLib {
	BitGroup::BitGroup(){
		this->init(4); /* default size */
	}

	BitGroup::BitGroup(int size){
		this->init(size);
	}

	int BitGroup::size(){
		return this->_group_size;
	}
	
	void BitGroup::copy(const BitGroup &original){
		int i;
		int size_of_int = 8 * sizeof(unsigned int);
		int nb_int = (original._group_size / size_of_int) + 1;
		this->init(original._group_size);	
		for (i=0; i < nb_int; i++){
			this->_bitgroup[i] = original._bitgroup[i];
		}
	}

	bool BitGroup::operator!=(const BitGroup &original){
		return ((*this)==original)?false:true;
	}

	bool BitGroup::operator==(const BitGroup &original){
		bool identity = true;
		if (this->_group_size == original._group_size){
			int size_of_int = 8 * sizeof(unsigned int);
			int nb_int = (original._group_size / size_of_int) + 1;
			for (int i=0; i < nb_int; i++){
				if (this->_bitgroup[i] != original._bitgroup[i]){
					// not the same content
					identity = false;
					break;
				}
			}
		} else {
			// not the same length
			identity=false;
		}
		return identity;
	}

	BitGroup BitGroup::operator=(BitGroup &original){
		this->copy(original);
		return (*this);
	}

	BitGroup BitGroup::operator=(const BitGroup &original){
		this->copy(original);
		return (*this);
	}

	BitGroup::BitGroup(const BitGroup &original){
		this->copy(original);
	}

	BitGroup::~BitGroup(){
		delete(this->_bitgroup);
	}
	
	void BitGroup::init(int size){
		int i;
		int size_of_int = 8 * sizeof(unsigned int);
		int nb_int = (size / size_of_int) + 1;
		this->_bitgroup = new unsigned int[nb_int];
		this->_group_size = size;
		for (i=0; i < nb_int; i++){
			this->_bitgroup[i] = 0;
		}
	}

	void BitGroup::setBitAt(int position, bool value){
		if (position > this->_group_size){
			throw eBitGroupOutOfBound();
		} else {
			int i;
			int size_of_int = 8 * sizeof(int);
			int index_of_int = position / size_of_int;
			int shift_of_int = position % size_of_int;
			unsigned int mask_fill;
			unsigned int mask_hole;
			mask_fill = 0;
			mask_hole = 0;
			for (i=0; i< size_of_int; i++){ 
				mask_fill = (mask_fill << 1) | 1;
				if (shift_of_int == i){
					mask_hole = ( mask_hole << 1 ) | 0;
				} else {
					mask_hole = ( mask_hole << 1 ) | 1;
				}
			}
			if (value){
				this->_bitgroup[index_of_int] = ((this->_bitgroup[index_of_int] ^ mask_fill) & mask_hole ) ^ mask_fill;
			} else {
				this->_bitgroup[index_of_int] = this->_bitgroup[index_of_int] & mask_hole;	
			}
			//printf("0x%u\n",this->_bitgroup[index_of_int]);
		}
	}

	bool BitGroup::getBitAt(int position){
		if (position > this->_group_size){
			//FIXME: throw exception 
			throw eBitGroupOutOfBound();
		} else {
			int i;
			int size_of_int =  8 * sizeof(unsigned int);
			int index_of_int = position / size_of_int;
			int shift_of_int = position % size_of_int;
			unsigned int reader = this->_bitgroup[index_of_int];
			if (shift_of_int > 0){ reader = reader << shift_of_int; reader = reader >> shift_of_int; }
			reader = reader >> (size_of_int -  shift_of_int - 1 ); 
			return reader;
		}
	}

	void BitGroup::setValue(unsigned long int value){
		unsigned long int val;
		val = value;
		if (val > this->maxValue()){
			throw eBitGroupOutOfRangeValue();
		} else {
			int position = (this->_group_size -1);
			while (val>0){
				this->setBitAt(position, val%2);
				val = val >> 1;
				position--;
			}
		}
	}
	unsigned long int BitGroup::maxValue(){
		return ((1 << this->_group_size)-1);
	}
	unsigned long int BitGroup::getValue(){
		unsigned long int result;
		result = 0;
		for (int i=0; i<this->_group_size; i++){
			result = (result << 1) | this->getBitAt(i);
		}
		return result;
	}

	std::string BitGroup::toString(){
		// faire un nouveau string
		std::string s;
		int i;
		for (i=0; i<this->_group_size; i++){
			s += (this->getBitAt(i)?"1":"0");
		}
		return s;
	}

}
